/* 
 * nre.c --
 *
 *	This file contains the commands for the nre package.
 *
 * Copyright (c) 1997 Darrel Schneider
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */
#include <tcl.h>
#include <string.h>
#include "nre.h"

static void	DupRegExpInternalRep _ANSI_ARGS_((Tcl_Obj *srcPtr,
						  Tcl_Obj *copyPtr));
static void	FreeRegExpInternalRep _ANSI_ARGS_((Tcl_Obj *rePtr));
static int	SetRegExpFromAny _ANSI_ARGS_((Tcl_Interp *interp,
					      Tcl_Obj *objPtr));
static void	UpdateStringOfRegExp _ANSI_ARGS_((Tcl_Obj *rePtr));
extern
#if defined(_WIN32) || defined(__WIN32__)
__declspec(dllexport)
#endif
int Nre_Init _ANSI_ARGS_((Tcl_Interp* interp));

/*
 * The structure below defines the list Tcl object type by means of
 * procedures that can be invoked by generic object code.
 */

static Tcl_ObjType tclRegExpType = {
    "regexp",				/* name */
    FreeRegExpInternalRep,		/* freeIntRepProc */
    DupRegExpInternalRep,	        /* dupIntRepProc */
    UpdateStringOfRegExp,		/* updateStringProc */
    SetRegExpFromAny			/* setFromAnyProc */
};

#define NRE_CACHE_SIZE 7

/* nrecache should really be part of the interp state in tclInt.h
 * Right now the cache is only used for the top-level interpreter.
 */
static Tcl_Obj* nrecache[NRE_CACHE_SIZE] = {NULL};

static void
addToNreCache(interp, objPtr)
    Tcl_Interp *interp;		/* Used for error reporting if not NULL. */
    Tcl_Obj *objPtr;		/* The object to consider adding. */
{
  int i;
  /* PRECONDITION: objPtr->typePtr == &tclRegExpType */

  if (Tcl_IsShared(objPtr)) {
    /* don't bother adding it since the user has probably already
     * cached it in a variable.
     */
    return;
  }
  if (Tcl_GetMaster(interp) != NULL) {
    /* This cache is only contains objects that belong to the top level
     * interpreter. That way we won't get in trouble by caching on object
     * of an interpreter that get blown away.
     * Since 'interp' had a master we know it was not the top level.
     */
    return;
  }
  
  if (nrecache[NRE_CACHE_SIZE-1] != NULL) {
    /* cache is full. So release the last element. */
    Tcl_DecrRefCount(nrecache[NRE_CACHE_SIZE-1]);
    nrecache[NRE_CACHE_SIZE-1] = NULL;
  }
  for (i = NRE_CACHE_SIZE-1; i > 0; i--) {
    nrecache[i] = nrecache[i-1];
  }
  nrecache[0] = objPtr;
  /* the cache will have a reference so increment it */
  Tcl_IncrRefCount(objPtr);
}

static int
findInNreCache(interp, objPtr, str, length)
    Tcl_Interp *interp;		/* Used for error reporting if not NULL. */
    Tcl_Obj *objPtr;		/* The object to find. */
    char* str;                  /* the object's string */
    int length;                 /* the object's string's length */
{
  int j;
  int idx = 0;
  Tcl_Obj* cacheObjPtr;

  while ((cacheObjPtr = nrecache[idx]) != NULL) {
    if (cacheObjPtr->typePtr != &tclRegExpType) {
      /* This guy is no longer a regexp object so remove it from the cache */
      Tcl_DecrRefCount(cacheObjPtr);
      for (j = idx; j < NRE_CACHE_SIZE-1; j++) {
	nrecache[j] = nrecache[j+1];
      }
      nrecache[NRE_CACHE_SIZE-1] = NULL;
      /* note we do not increment idx since everything was moved down */
    } else {
      /* Check to see if the cacheObj is the same as the current guy */
      int cacheStrLen;
      char* cacheStr = Tcl_GetStringFromObj(cacheObjPtr, &cacheStrLen);
      if (cacheStrLen == length) {
	if (0 == memcmp(str, cacheStr, length)) {
	  /* move this guy to slot 0 since he was used */
	  if (idx != 0) {
	    for (j = idx; j > 0; j--) {
	      nrecache[j] = nrecache[j-1];
	    }
	    nrecache[0] = cacheObjPtr;
	  }
	  DupRegExpInternalRep(cacheObjPtr, objPtr);
	  return 1;
	}
      }
      idx++;
      if (idx == NRE_CACHE_SIZE) {
	return 0;
      }
    }
  }
  return 0;
}
/*
 *----------------------------------------------------------------------
 *
 * FreeRegExpInternalRep --
 *
 *	Deallocate the storage associated with a regexp object's internal
 *	representation.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Frees rePtr's List* internal representation and sets rePtr's
 *	internalRep.otherValuePtr to NULL.
 *
 *----------------------------------------------------------------------
 */

static void
FreeRegExpInternalRep(rePtr)
    Tcl_Obj *rePtr;		/* RegExp object with internal rep to free. */
{
    ckfree((char *) rePtr->internalRep.otherValuePtr);
    rePtr->internalRep.otherValuePtr = NULL;
}

/*
 *----------------------------------------------------------------------
 *
 * DupRegExpInternalRep --
 *
 *	Initialize the internal representation of a regexp Tcl_Obj to a
 *	copy of the internal representation of an existing regexp object. 
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	"srcPtr"s regexp internal rep pointer should not be NULL and we assume
 *	it is not NULL. We set "copyPtr"s internal rep to a pointer to a
 *	newly allocated Tcl_RegExp structure that, in turn, is set to contain
 *      a copy of srcPtr.
 *
 *----------------------------------------------------------------------
 */

static void
DupRegExpInternalRep(srcPtr, copyPtr)
    Tcl_Obj *srcPtr;		/* Object with internal rep to copy. */
    Tcl_Obj *copyPtr;		/* Object with internal rep to set. */
{
    regexp *srcRepPtr = (regexp*) srcPtr->internalRep.otherValuePtr;
    regexp *cpyRepPtr;
    int psize = srcRepPtr->regexpSize;

    cpyRepPtr = (regexp*) ckalloc(psize);
    memcpy(cpyRepPtr, srcRepPtr, psize);

    /* fixup the fields that are pointers */
    if (cpyRepPtr->regstartnode != NULL) {
      cpyRepPtr->regstartnode -= (unsigned long)srcRepPtr;
      cpyRepPtr->regstartnode += (unsigned long)cpyRepPtr;
    }
    if (cpyRepPtr->regmust != NULL) {
      cpyRepPtr->regmust -= (unsigned long)srcRepPtr;
      cpyRepPtr->regmust += (unsigned long)cpyRepPtr;
    }
    if (cpyRepPtr->regclasses != NULL) {
      cpyRepPtr->regclasses -= (unsigned long)srcRepPtr;
      cpyRepPtr->regclasses += (unsigned long)cpyRepPtr;
    }
    
    copyPtr->internalRep.otherValuePtr = (VOID *) cpyRepPtr;
    copyPtr->typePtr = &tclRegExpType;
}

/*
 *----------------------------------------------------------------------
 *
 * SetRegExpFromAny --
 *
 *	Attempt to generate a regexp internal form for the Tcl object
 *	"objPtr".
 *
 * Results:
 *	The return value is TCL_OK or TCL_ERROR. If an error occurs during
 *	conversion, an error message is left in the interpreter's result
 *	unless "interp" is NULL.
 *
 * Side effects:
 *	If no error occurs, a regexp is stored as "objPtr"s internal
 *	representation.
 *
 *----------------------------------------------------------------------
 */

static int
SetRegExpFromAny(interp, objPtr)
    Tcl_Interp *interp;		/* Used for error reporting if not NULL. */
    Tcl_Obj *objPtr;		/* The object to convert. */
{
    regexp *rePtr;
    Tcl_ObjType *oldTypePtr = objPtr->typePtr;
    char *string;
    int length;

    if (oldTypePtr == &tclRegExpType) {
      /* Already is a regexp object */
      return TCL_OK;
    }

    /*
     * Get the string representation. Make it up-to-date if necessary.
     */

    string = Tcl_GetStringFromObj(objPtr, &length);

    /*
     * Free the old internalRep before setting the new one.
     * Note we do this after Tcl_GetStringFromObj.
     */
    if ((oldTypePtr != NULL) && (oldTypePtr->freeIntRepProc != NULL)) {
	oldTypePtr->freeIntRepProc(objPtr);
        objPtr->typePtr = NULL;
    }

    if (!findInNreCache(interp, objPtr, string, length)) {
      TclNewRegError((char *) NULL);
      rePtr = TclNewRegCompSz(string, length, 0);
      if (TclNewGetRegError() != NULL) {
	Tcl_AppendResult(interp,
			 "couldn't compile regular expression pattern: ",
			 TclNewGetRegError(), (char *) NULL);
	return TCL_ERROR;
      }

      objPtr->internalRep.otherValuePtr = (VOID *) rePtr;
      objPtr->typePtr = &tclRegExpType;
      addToNreCache(interp, objPtr);
    }
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * UpdateStringOfRegExp --
 *
 *	Update the string representation for a regexp object.
 *	Note: regexp object's are invariant so whatever string
 *      they had when created or dupped is ok. So this function is a noop.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

static void
UpdateStringOfRegExp(rePtr)
    Tcl_Obj *rePtr;
{
    if (rePtr->bytes == NULL) {
#if 0 /* use this code if nre is moved into the core */
	rePtr->bytes = tclEmptyStringRep;
#else
	rePtr->bytes = (char *) ckalloc((unsigned) 1);
	rePtr->bytes[0] = '\0';
#endif
	rePtr->length = 0;
    }
    return;
  /* do nothing */
}


static Tcl_Obj*
Tcl_RegExpCompileObj(interp, pattern, noCase)
    Tcl_Interp *interp;			/* For use in error reporting. */
    Tcl_Obj *pattern;			/* String for which to produce
					 * compiled regular expression. */
    int noCase;                         /* if 1 case insensitive */
{
  /*
   * If noCase is 1 then the user explicity asked for case insensitive.
   * Otherwise it is up to the pattern's contents (i.e. (?i) switch).
   */
  if (pattern->typePtr == &tclRegExpType) {
    int allDone = 1;
    if (noCase) {
      /* Need to check pattern's state and see if it is noCase */
      regexp *rePtr = (regexp*)pattern->internalRep.otherValuePtr;
      if (!(rePtr->regflags & REGEXP_NOCASE)) {
	/* Need to recompile
	 * NYI: should check if it is shared and if so create a new object!
	 */
	allDone = 0;
      }
    }
    if (allDone) {
      return pattern;
    }
  }
  if (noCase) {
    /* inline version of setRegExpFromAny.
     * Need this hack so we can specify noCase switch.
     */
    regexp *rePtr;
    Tcl_ObjType *oldTypePtr = pattern->typePtr;
    char *string;
    int length;

    string = Tcl_GetStringFromObj(pattern, &length);

    /*
     * Free the old internalRep before setting the new one.
     * Note we do this after Tcl_GetStringFromObj.
     */
    if ((oldTypePtr != NULL) && (oldTypePtr->freeIntRepProc != NULL)) {
	oldTypePtr->freeIntRepProc(pattern);
        pattern->typePtr = NULL;
    }

    TclNewRegError((char *) NULL);
    rePtr = TclNewRegCompSz(string, length, noCase);
    if (TclNewGetRegError() != NULL) {
	Tcl_AppendResult(interp,
	    "couldn't compile regular expression pattern: ",
	    TclNewGetRegError(), (char *) NULL);
	return NULL;
    }

    pattern->internalRep.otherValuePtr = (VOID *) rePtr;
    pattern->typePtr = &tclRegExpType;
    return pattern;
  } else {
    if (Tcl_ConvertToType(interp, pattern, &tclRegExpType) == TCL_OK) {
      return pattern;
    } else {
      return NULL;
    }
  }
}

static int
Tcl_RegExpExecObj(interp, reObj, string, start, stringLength, resptr)
    Tcl_Interp *interp;		/* Interpreter to use for error reporting. */
    Tcl_Obj* reObj;		/* Compiled regular expression;  must have
				 * been returned by previous call to
				 * Tcl_RegExpCompileObj. */
    char *string;		/* String against which to match re. */
    char *start;		/* If string is part of a larger string,
				 * this identifies beginning of larger
				 * string, so that "^" won't match. */
    int stringLength;
    Tcl_RegExpExecResult *resptr;
{
    int match;
    regexp *regexpPtr = (regexp*)reObj->internalRep.otherValuePtr;

    if (reObj->typePtr != &tclRegExpType) {
	Tcl_ResetResult(interp);
	Tcl_AppendResult(interp, "expect a regular expression", (char *) NULL);
	return -1;
    }

    TclNewRegError((char *) NULL);
    match = TclNewRegExecSz(regexpPtr, string, start, stringLength, resptr);
    if (TclNewGetRegError() != NULL) {
	Tcl_ResetResult(interp);
	Tcl_AppendResult(interp, "error while matching regular expression: ",
		TclNewGetRegError(), (char *) NULL);
	return -1;
    }
    return match;
}

static void
Tcl_NewRegExpRange(re, resptr, index, startPtr, endPtr)
    Tcl_RegExp re;		/* Compiled regular expression that has
				 * been passed to Tcl_RegExpExec. */
    Tcl_RegExpExecResult *resptr;           /* Result from Tcl_RegExpExec */
    int index;			/* 0 means give the range of the entire
				 * match, > 0 means give the range of
				 * a matching subrange.  Must be no greater
				 * than NSUBEXP. */
    char **startPtr;		/* Store address of first character in
				 * (sub-) range here. */
    char **endPtr;		/* Store address of character just after last
				 * in (sub-) range here. */
{
    regexp *regexpPtr = (regexp *) re;

    if (index >= regexpPtr->subexpCount) {
	*startPtr = *endPtr = NULL;
    } else {
	*startPtr = resptr->startp[index];
	*endPtr = resptr->endp[index];
    }
}

/* Object form of Tcl_RegExpRange.
 */
static void
Tcl_RegExpRangeObj(reObj, resptr, index, startPtr, endPtr)
    Tcl_Obj* reObj;		/* Compiled regular expression that has
				 * been passed to Tcl_RegExpExec. */
    Tcl_RegExpExecResult *resptr;           /* Result from Tcl_RegExpExec */
    int index;			/* 0 means give the range of the entire
				 * match, > 0 means give the range of
				 * a matching subrange.  Must be no greater
				 * than NSUBEXP. */
    char **startPtr;		/* Store address of first character in
				 * (sub-) range here. */
    char **endPtr;		/* Store address of character just after last
				 * in (sub-) range here. */
{
  Tcl_NewRegExpRange((Tcl_RegExp)reObj->internalRep.otherValuePtr,
		     resptr, index, startPtr, endPtr);
}

static int
Tcl_RegExpMatchObj(interp, string, stringLength, reObj)
    Tcl_Interp *interp;		/* Used for error reporting. */
    char *string;		/* String. */
    int stringLength;
    Tcl_Obj* reObj;
{
    Tcl_RegExpExecResult reres;
    if (Tcl_ConvertToType(interp, reObj, &tclRegExpType) != TCL_OK) {
      return -1;
    }
    return Tcl_RegExpExecObj(interp, reObj, string, string, stringLength,
			     &reres);
}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_NewLsearchObjCmd --
 *
 *	This procedure is invoked to process the "nrelsearch" Tcl command.
 *	See the user documentation for details on what it does.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	See the user documentation.
 *
 *----------------------------------------------------------------------
 */

static int
Tcl_NewLsearchObjCmd(clientData, interp, objc, objv)
    ClientData clientData;	/* Not used. */
    Tcl_Interp *interp;		/* Current interpreter. */
    int objc;			/* Number of arguments. */
    Tcl_Obj *CONST objv[];	/* Argument values. */
{
#define EXACT	0
#define GLOB	1
#define REGEXP	2
    char *bytes, *patternBytes;
    Tcl_Obj* patternObj;
    int i, match, mode, index, result, listLen, length, elemLen;
    Tcl_Obj **elemPtrs;
    static char *switches[] =
	    {"-exact", "-glob", "-regexp", (char *) NULL};

    mode = GLOB;
    if (objc == 4) {
	if (Tcl_GetIndexFromObj(interp, objv[1], switches,
		"search mode", 0, &mode) != TCL_OK) {
	    return TCL_ERROR;
	}
    } else if (objc != 3) {
	Tcl_WrongNumArgs(interp, 1, objv, "?mode? list pattern");
	return TCL_ERROR;
    }

    /*
     * Make sure the list argument is a list object and get its length and
     * a pointer to its array of element pointers.
     */

    result = Tcl_ListObjGetElements(interp, objv[objc-2], &listLen, &elemPtrs);
    if (result != TCL_OK) {
	return result;
    }

    patternObj = objv[objc-1];
    if (mode != REGEXP) {
      patternBytes = Tcl_GetStringFromObj(patternObj, &length);
    }
 
    index = -1;
    for (i = 0; i < listLen; i++) {
	match = 0;
	bytes = Tcl_GetStringFromObj(elemPtrs[i], &elemLen);
	switch (mode) {
	    case EXACT:
		if (length == elemLen) {
		    match = (memcmp(bytes, patternBytes,
			    (size_t) length) == 0);
		}
		break;
	    case GLOB:
		/*
		 * WARNING: will not work with data containing NULLs.
		 */
		match = Tcl_StringMatch(bytes, patternBytes);
		break;
	    case REGEXP:
		match = Tcl_RegExpMatchObj(interp, bytes, elemLen, patternObj);
		if (match < 0) {
		    return TCL_ERROR;
		}
		break;
	}
	if (match) {
	    index = i;
	    break;
	}
    }

    Tcl_SetIntObj(Tcl_GetObjResult(interp), index);
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_NewRegexpObjCmd --
 *
 *	This procedure is invoked to process the "nrematch" Tcl command.
 *	See the user documentation for details on what it does.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	See the user documentation.
 *
 *----------------------------------------------------------------------
 */

	/* ARGSUSED */
static int
Tcl_NewRegexpObjCmd(dummy, interp, objc, objv)
    ClientData dummy;			/* Not used. */
    Tcl_Interp *interp;			/* Current interpreter. */
    int objc;				/* Number of arguments. */
    Tcl_Obj *CONST objv[];		/* Argument objects. */
{
    int noCase = 0;
    int indices = 0;
    int doAll = 0;
    Tcl_Obj * patternObj;
    char *start, *end;
    char *string, *stringEnd;
    int stringLength;
    char *argStr;
    Tcl_Obj *CONST *argPtr;
    int numMatches = 0;
    int match = 0;			/* Initialization needed only to
					 * prevent compiler warning. */
    int i;
    int nameLength;
    Tcl_Obj* emptyObj = NULL;
    Tcl_Obj* varObjs[NEWNSUBEXP+1];
    int maxVarObjs;
    int code;
    static char *switches[] =
	    {"-all", "-indices", "-nocase", "--", (char *) NULL};
    int switchObjc, index, length;
    Tcl_Obj *CONST *switchObjv;
    enum regexpSwitches {ALL_SWITCH, INDICES_SWITCH, NOCASE_SWITCH, END_SWITCH};
    Tcl_RegExpExecResult reRes;

    switchObjc = objc-1;
    switchObjv = objv+1;

    while (switchObjc > 0) {
	argStr = Tcl_GetStringFromObj(switchObjv[0], &length);
	if (*argStr != '-') {
	  break;
	}
	if (Tcl_GetIndexFromObj(interp, switchObjv[0], switches,
		"option", TCL_EXACT, &index) != TCL_OK) {
	    return TCL_ERROR;
	}
	switch (index) {
	    case ALL_SWITCH:
		doAll = 1;
		break;
	    case INDICES_SWITCH:
		indices = 1;
		break;
	    case NOCASE_SWITCH:
		noCase = 1;
		break;
	    case END_SWITCH:			/* -- */
		switchObjc--;
		switchObjv++;
		goto doneWithSwitches;
	}
	switchObjc--;
	switchObjv++;
    }

    doneWithSwitches:
    if (switchObjc < 2) {
	Tcl_WrongNumArgs(interp, 1, objv,
	  "?switches? exp string ?matchVar? ?subMatchVar subMatchVar ...?");
	return TCL_ERROR;
    }
    objc = switchObjc - 2;
    /* objc is now the number of variable names to store matches in */
    argPtr = switchObjv;

    patternObj = (Tcl_Obj*)argPtr[0];
    string = Tcl_GetStringFromObj(argPtr[1], &stringLength);
    stringEnd = string + stringLength;
    patternObj = Tcl_RegExpCompileObj(interp, patternObj, noCase);
    code = TCL_ERROR;
    maxVarObjs = 0;
    if (patternObj != NULL) {
      numMatches = 0;
      maxVarObjs = objc;
      if (maxVarObjs > NEWNSUBEXP+1) {
	maxVarObjs = NEWNSUBEXP+1;
      }
      if (maxVarObjs > 0) {
	if (indices) {
	  if (doAll) {
	    emptyObj = Tcl_NewIntObj(-1);
	  } else {
	    Tcl_Obj* objArray[2];
	    objArray[0] = Tcl_NewIntObj(-1);
	    objArray[1] = objArray[0];
	    emptyObj = Tcl_NewListObj(2, objArray);
	  }
	} else {
	  emptyObj = Tcl_NewObj();
	}
	Tcl_IncrRefCount(emptyObj);
      }
      if (doAll) {
	char* p;
	for (i = 0; i < maxVarObjs; i++) {
	  Tcl_GetStringFromObj((Tcl_Obj*)argPtr[i+2], &nameLength);
	  if (nameLength != 0) {
	    varObjs[i] = Tcl_NewObj();
	    Tcl_IncrRefCount(varObjs[i]);
	  } else {
	    varObjs[i] = NULL;
	  }
	}
	code = TCL_OK;
	for (p = string; p <= stringEnd; ) {
	  Tcl_Obj** storePtr;
	  match = Tcl_RegExpExecObj(interp, patternObj,
				    p, string, stringEnd - p, &reRes);
	  if (match < 0) {
	    code = TCL_ERROR;
	    break;
	  }
	  if (!match) {
	    break;
	  }
	  numMatches++;
	  storePtr = varObjs;
	  for (i = 0; i < maxVarObjs; i++) {
	    if (*storePtr != NULL) {
	      Tcl_Obj* valPtr;
	      Tcl_RegExpRangeObj(patternObj, &reRes, i, &start, &end);
	      if (indices) {
		if (start == NULL) {
		  Tcl_ListObjAppendElement(interp, *storePtr, emptyObj);
		  Tcl_ListObjAppendElement(interp, *storePtr, emptyObj);
		} else {
		  Tcl_ListObjAppendElement(interp, *storePtr,
					   Tcl_NewIntObj((int)(start - string)));
		  Tcl_ListObjAppendElement(interp, *storePtr,
					   Tcl_NewIntObj((int)(end - string - 1)));
		}
	      } else {
		if (start == NULL) {
		  valPtr = emptyObj;
		} else {
		  int length = (end - start);
		  if (length == 0) {
		    valPtr = emptyObj;
		  } else {
		    valPtr = Tcl_NewStringObj(start, length);
		  }
		}
		Tcl_ListObjAppendElement(interp, *storePtr, valPtr);
	      }
	    }
	    storePtr++;
	  }
	  Tcl_RegExpRangeObj(patternObj, &reRes, 0, &start, &end);
	  if (end == p) {
	    /*
	     * Always consume at least one character of the input string
	     * in order to prevent infinite loops.
	     */
	    p = end + 1;
	  } else {
	    p = end;
	  }
	} /* end for */
      } else {
	for (i = 0; i < maxVarObjs; i++) {
	  varObjs[i] = NULL;
	}
	match = Tcl_RegExpExecObj(interp, patternObj,
				  string, string, stringEnd - string, &reRes);
	if (match >= 0) {
	  code = TCL_OK;
	  if (match > 0) {
	    numMatches++;
	    for (i = 0; i < maxVarObjs; i++) {
	      Tcl_GetStringFromObj((Tcl_Obj*)argPtr[i+2], &nameLength);
	      if (nameLength != 0) {
		Tcl_RegExpRangeObj(patternObj, &reRes, i, &start, &end);
		if (start == NULL) {
		  varObjs[i] = emptyObj;
		} else {
		  if (indices) {
		    Tcl_Obj* objArray[2];
		    objArray[0] = Tcl_NewIntObj((int)(start - string));
		    objArray[1] = Tcl_NewIntObj((int)(end - string - 1));
		    varObjs[i] = Tcl_NewListObj(2, objArray);
		  } else {
		    int length = (end - start);
		    if (length == 0) {
		      varObjs[i] = emptyObj;
		    } else {
		      varObjs[i] = Tcl_NewStringObj(start, length);
		    }
		  }
		}
		Tcl_IncrRefCount(varObjs[i]);
	      }
	    }
	  }
	}
      }
    }
    if (code == TCL_OK && numMatches > 0) {
      /*
       * If additional variable names have been specified, return
       * index information in those variables.
       */

      for (i = 0; i < objc; i++) {
	Tcl_Obj* nameObj;
	Tcl_Obj* valueObj;
	int j = i;
	nameObj = (Tcl_Obj*)argPtr[i+2];
	Tcl_GetStringFromObj(nameObj, &nameLength);
	if (nameLength != 0) {
	  if (j >= maxVarObjs) {
	    j = maxVarObjs-1;
	  }
	  valueObj = varObjs[j];
	  if (valueObj == NULL) {
	    valueObj = emptyObj;
	  }
	  if (!Tcl_ObjSetVar2(interp, nameObj, NULL, valueObj, TCL_PARSE_PART1)) {
	    Tcl_AppendResult(interp, "couldn't set variable \"",
			     Tcl_GetStringFromObj(nameObj, (int*)NULL),
			     "\"", (char *) NULL);
	    code = TCL_ERROR;
	    break;
	  }
	}
      }
    }
    if (code == TCL_OK) {
      Tcl_SetObjResult(interp, Tcl_NewIntObj(numMatches));
    }
    /* Free objects */
    for (i = 0; i < maxVarObjs; i++) {
      if (varObjs[i] != NULL) {
	Tcl_DecrRefCount(varObjs[i]);
      }

    }
    if (emptyObj != NULL) {
      Tcl_DecrRefCount(emptyObj);
    }
    return code;
}

static int
Tcl_NewRegDumpObjCmd(dummy, interp, objc, objv)
    ClientData dummy;			/* Not used. */
    Tcl_Interp *interp;			/* Current interpreter. */
    int objc;				/* Number of arguments. */
    Tcl_Obj *CONST objv[];		/* Argument objects. */
{
    char buff[80];
    int alreadyRegexp;
    Tcl_Obj * patternObj;
    Tcl_Obj *resObjPtr;

    if (objc != 2) {
	Tcl_WrongNumArgs(interp, 1, objv, "exp");
	return TCL_ERROR;
    }
    patternObj = (Tcl_Obj*)objv[1];
    alreadyRegexp = (patternObj->typePtr == &tclRegExpType);
    patternObj = Tcl_RegExpCompileObj(interp, patternObj, 0);
    if (patternObj == NULL) {
      return TCL_ERROR;
    }
    if (TclNewRegDump(interp, (regexp*)patternObj->internalRep.otherValuePtr)
	== TCL_ERROR) {
      return TCL_ERROR;
    }
    resObjPtr = Tcl_GetObjResult(interp);
    sprintf(buff, "exp's object %p%s refcount: %d",
	    patternObj,
	    alreadyRegexp ? "" : " needed compilation",
	    patternObj->refCount);

    return Tcl_ListObjAppendElement(interp, resObjPtr,
				    Tcl_NewStringObj(buff, -1));
}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_NewRegsubCmd --
 *
 *	This procedure is invoked to process the "nresub" Tcl command.
 *	See the user documentation for details on what it does.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	See the user documentation.
 *
 *----------------------------------------------------------------------
 */
	/* ARGSUSED */
static int
Tcl_NewRegsubObjCmd(dummy, interp, objc, objv)
    ClientData dummy;			/* Not used. */
    Tcl_Interp *interp;			/* Current interpreter. */
    int objc;				/* Number of arguments. */
    Tcl_Obj *CONST objv[];		/* Argument objects. */
{
    int noCase = 0, all = 0;
    Tcl_Obj * patternObj;
    char *p, *firstChar;
    int match, numMatches;
    char *start, *end, *subStart, *subEnd;
    register char *src, c;
    Tcl_Obj *subObj = NULL;
    char *string, *stringEnd, *subSpec, *subSpecEnd;
    int stringLength, subSpecLength;
    Tcl_Obj *CONST *argPtr;
    static char *switches[] =
	    {"-all", "-nocase", "-eval", "--", (char *) NULL};
    int switchObjc, index, length;
    Tcl_Obj *CONST *switchObjv;
    enum regsubSwitches {ALL_SWITCH, NOCASE_SWITCH, EVAL_SWITCH, END_SWITCH};
    Tcl_RegExpExecResult reRes;
    int eval = 0;
    Tcl_Obj *subSpecObj = NULL;
    Tcl_Obj *evalValueObj = NULL;
    Tcl_Obj *matchStringArray[NEWNSUBEXP];
    int numMatchStrings;
    int evalSetFlags;
    int i;
    static Tcl_Obj *emptyStrObj = NULL;
    int doContinue;
    Tcl_Obj *evalNameObj = NULL;

    switchObjc = objc-1;
    switchObjv = objv+1;

    while (switchObjc > 0) {
	char *argStr = Tcl_GetStringFromObj(switchObjv[0], &length);
	if (*argStr != '-') {
	  break;
	}
	if (Tcl_GetIndexFromObj(interp, switchObjv[0], switches,
		"option", TCL_EXACT, &index) != TCL_OK) {
	    return TCL_ERROR;
	}
	switch (index) {
	    case ALL_SWITCH:
		all = 1;
		break;
	    case NOCASE_SWITCH:
		noCase = 1;
		break;
	    case EVAL_SWITCH:
		eval = 1;
		break;
	    case END_SWITCH:			/* -- */
		switchObjc--;
		switchObjv++;
		goto doneWithSwitches;
	}
	switchObjc--;
	switchObjv++;
    }

    doneWithSwitches:
    if (eval) {
      if (switchObjc != 5) {
	Tcl_WrongNumArgs(interp, 1, objv,
			 "-eval ?switches? exp string subSpec varName evalName");
	return TCL_ERROR;
      }
    } else {
      if (switchObjc != 4) {
	Tcl_WrongNumArgs(interp, 1, objv,
			 "?switches? exp string subSpec varName");
	return TCL_ERROR;
      }
    }
    objc = switchObjc;
    argPtr = switchObjv;

    patternObj = (Tcl_Obj*)argPtr[0];
    string = Tcl_GetStringFromObj(argPtr[1], &stringLength);
    stringEnd = string + stringLength;
    if (eval) {
      subSpecObj = argPtr[2];
      evalNameObj = argPtr[4];
    } else {
      subSpec = Tcl_GetStringFromObj(argPtr[2], &subSpecLength);
      subSpecEnd = subSpec + subSpecLength;
    }

    patternObj = Tcl_RegExpCompileObj(interp, patternObj, noCase);
    if (patternObj == NULL) {
      return TCL_ERROR;
    }

    subObj = Tcl_NewObj();
    if (eval) {
      /* Determine the size of the match list. We know the size of the list
       * will always be the number of subexpression in the pattern.
       */
      char *evalName;
      int evalNameLen;
      regexp* rePtr = (regexp*)patternObj->internalRep.otherValuePtr;
      numMatchStrings = rePtr->subexpCount;
      evalSetFlags = TCL_LEAVE_ERR_MSG;
      /* Figure out if we need to do a PARSE_PART1 every time we set the var */
      evalName = Tcl_GetStringFromObj(evalNameObj, &evalNameLen);
      if (memchr(evalName, '(', evalNameLen) != NULL) {
	/* The evalName contains a '(' so its probably an array name */
	evalSetFlags |= TCL_PARSE_PART1;
      }
    }
    /*
     * The following loop is to handle multiple matches within the
     * same source string;  each iteration handles one match and its
     * corresponding substitution.  If "-all" hasn't been specified
     * then the loop body only gets executed once.
     */

    numMatches = 0;
    for (p = string; p <= stringEnd; ) {
      doContinue = 0;
      match = Tcl_RegExpExecObj(interp, patternObj, p, string, stringEnd - p,
				&reRes);
      if (match < 0) {
	Tcl_DecrRefCount(subObj);
	return TCL_ERROR;
      }
      if (!match) {
	break;
      }
      numMatches += 1;

      /*
       * Copy the portion of the source string before the match to the
       * result variable.
       */

      Tcl_RegExpRangeObj(patternObj, &reRes, 0, &start, &end);
      Tcl_AppendToObj(subObj, p, start - p);
    
      if (eval) {
	int evalres;
	int alldone = 0;
	/* Create strings for all the matched expressions */
	for (i = 0; i < numMatchStrings; i++) {
	  int len;
	  Tcl_RegExpRangeObj(patternObj, &reRes, i, &start, &end);
	  len = end - start;
	  if (len > 0) {
	    matchStringArray[i] = Tcl_NewStringObj(start, len);
	  } else {
	    if (emptyStrObj == NULL) {
	      emptyStrObj = Tcl_NewObj();
	    }
	    matchStringArray[i] = emptyStrObj;
	  }
	}
	if (numMatchStrings == 1) {
	  /* no subexpressions so don't bother with a list */
	  evalValueObj = matchStringArray[0];
	} else {
	  evalValueObj = Tcl_NewListObj(numMatchStrings, matchStringArray);
	}
	if (!Tcl_ObjSetVar2(interp, evalNameObj, NULL, evalValueObj,
			    evalSetFlags)) {
	  Tcl_DecrRefCount(subObj);
	  Tcl_DecrRefCount(evalValueObj);
	  return TCL_ERROR;
	}
	/* Now evaluate the subSpec code which can have references to
	 * the evalName variable.
	 */
	evalres = Tcl_EvalObj(interp, subSpecObj);
	switch (evalres) {
	case TCL_OK: {
	  int resLength;
	  Tcl_Obj* resObj = Tcl_GetObjResult(interp);
	  char* resStr = Tcl_GetStringFromObj(resObj, &resLength);
	  /* TCL_OK means to take the result string and append it. */
	  Tcl_AppendToObj(subObj, resStr, resLength);
	  break;
	}
	case TCL_ERROR:
	  /* TCL_ERROR means that something went wrong. */
	  /* Need to set a special error message? */
	  Tcl_DecrRefCount(subObj);
	  return TCL_ERROR;
	case TCL_BREAK:
	  /* TCL_BREAK means to act as if no match occurred and quit. */
	  numMatches -= 1;
	  p = start;
	  alldone = 1;
	  break;
	case TCL_CONTINUE:
	  /* TCL_CONTINUE means to act as if no match occurred
	   * and try again at the next position.
	   */
	  numMatches -= 1;
	  p = start;
	  end = start;
	  doContinue = 1;
	  break;
	case TCL_RETURN:
	default:
	  /* TCL_RETURN or anything else means to clean up and pass the
	   * code back to the caller.
	   * Note that in this case the result variable is not set.
	   */
	  Tcl_DecrRefCount(subObj);
	  return evalres;
	}
	if (alldone) {
	  break;
	}
      } else {
	/*
	 * Append the subSpec argument to the variable, making appropriate
	 * substitutions.  This code is a bit hairy because of the backslash
	 * conventions and because the code saves up ranges of characters in
	 * subSpec to reduce the number of calls to Tcl_SetVar.
	 */
	for (src = firstChar = subSpec; src != subSpecEnd; src++) {
	  int index;

	  c = *src;
	  if (c == '&') {
	    index = 0;
	  } else if (c == '\\') {
	    c = src[1];
	    if ((c >= '0') && (c <= '9')) {
	      index = c - '0';
	    } else if ((c == '\\') || (c == '&')) {
	      *src = c;
	      Tcl_AppendToObj(subObj, firstChar, (src - firstChar) + 1);
	      *src = '\\';
	      firstChar = src+2;
	      src++;
	      continue;
	    } else {
	      continue;
	    }
	  } else {
	    continue;
	  }
	  if (firstChar != src) {
	    /* append from 'firstChar' upto, but not including 'src' */
	    Tcl_AppendToObj(subObj, firstChar, src - firstChar);
	  }
	  Tcl_RegExpRangeObj(patternObj, &reRes, index, &subStart, &subEnd);
	  if ((subStart != NULL) && (subEnd != NULL)) {
	    Tcl_AppendToObj(subObj, subStart, subEnd - subStart);
	  }
	  if (*src == '\\') {
	    src++;
	  }
	  firstChar = src+1;
	}
      }
      if (!eval) {
	if (firstChar != src) {
	  Tcl_AppendToObj(subObj, firstChar, subSpecEnd - firstChar);
	}
      }
      if (end == p) {

	/*
	 * Always consume at least one character of the input string
	 * in order to prevent infinite loops.
	 */

	if (p < stringEnd) {
	  Tcl_AppendToObj(subObj, p, 1);
	}
	p++;
      } else {
	p = end;
      }
      if (!all && !doContinue) {
	break;
      }
    }

    /*
     * Copy the portion of the source string after the last match to the
     * result variable.
     */
    if ((p < stringEnd)) {
      Tcl_AppendToObj(subObj, p, stringEnd - p);
    }
    if (!Tcl_ObjSetVar2(interp, argPtr[3], NULL, subObj, TCL_PARSE_PART1)) {
	Tcl_AppendResult(interp,
			 "couldn't set variable \"",
			 Tcl_GetStringFromObj(argPtr[3], (int*)NULL),
			 "\"",(char *) NULL);
	Tcl_DecrRefCount(subObj);
	return TCL_ERROR;
    } else {
	Tcl_SetObjResult(interp, Tcl_NewIntObj(numMatches));
	return TCL_OK;
    }
}


/*
 *----------------------------------------------------------------------
 *
 * Tcl_NewSwitchObjCmd --
 *
 *	This object-based procedure is invoked to process the "nreswitch" Tcl
 *	command. See the user documentation for details on what it does.
 *
 * Results:
 *	A standard Tcl object result.
 *
 * Side effects:
 *	See the user documentation.
 *
 *----------------------------------------------------------------------
 */

	/* ARGSUSED */
static int
Tcl_NewSwitchObjCmd(dummy, interp, objc, objv)
    ClientData dummy;		/* Not used. */
    Tcl_Interp *interp;		/* Current interpreter. */
    int objc;			/* Number of arguments. */
    Tcl_Obj *CONST objv[];	/* Argument objects. */
{
#define EXACT	0
#define GLOB	1
#define REGEXP	2
    int switchObjc, index;
    Tcl_Obj *CONST *switchObjv;
    Tcl_Obj *patternObj, *bodyObj;
    char *string, *pattern, *body;
    int splitObjs, length, patternLen, i, code, mode, matched, bodyIdx;
    static char *switches[] =
	    {"-exact", "-glob", "-regexp", "--", (char *) NULL};

    switchObjc = objc-1;
    switchObjv = objv+1;
    mode = EXACT;
    
    string = Tcl_GetStringFromObj(switchObjv[0], &length);
    while ((switchObjc > 0) && (*string == '-')) {
	if (Tcl_GetIndexFromObj(interp, switchObjv[0], switches,
				"option", 0, &index) != TCL_OK) {
	    return TCL_ERROR;
	}
	switch (index) {
	    case 0:			/* -exact */
		mode = EXACT;
		break;
	    case 1:			/* -glob */
		mode = GLOB;
		break;
	    case 2:			/* -regexp */
		mode = REGEXP;
		break;
	    case 3:			/* -- */
		switchObjc--;
		switchObjv++;
		goto doneWithSwitches;
	}
	switchObjc--;
	switchObjv++;
	string = Tcl_GetStringFromObj(switchObjv[0], &length);
    }

    doneWithSwitches:
    if (switchObjc < 2) {
	Tcl_WrongNumArgs(interp, 1, objv,
		"?switches? string pattern body ... ?default body?");
	return TCL_ERROR;
    }
    
    string = Tcl_GetStringFromObj(switchObjv[0], &length);
    switchObjc--;
    switchObjv++;

    /*
     * If all of the pattern/command pairs are lumped into a single
     * argument, split them out again.
     */

    splitObjs = 0;
    if (switchObjc == 1) {
	code = Tcl_ListObjLength(interp, switchObjv[0], &switchObjc);
	if (code != TCL_OK) {
	    return code;
	}
	splitObjs = 1;
    }

    for (i = 0;  i < switchObjc;  i += 2) {
	if (i == (switchObjc-1)) {
	    Tcl_ResetResult(interp);
	    Tcl_AppendToObj(Tcl_GetObjResult(interp),
	            "extra switch pattern with no body", -1);
	    code = TCL_ERROR;
	    goto done;
	}

	/*
	 * See if the pattern matches the string.
	 */

	if (splitObjs) {
	    code = Tcl_ListObjIndex(interp, switchObjv[0], i, &patternObj);
	    if (code != TCL_OK) {
		return code;
	    }
	} else {
	    patternObj = switchObjv[i];
	}
	pattern = Tcl_GetStringFromObj(patternObj, &patternLen);

	matched = 0;
	if ((*pattern == 'd') && (i == switchObjc-2)
		&& (strcmp(pattern, "default") == 0)) {
	    matched = 1;
	} else {
	    switch (mode) {
		case EXACT:
		    matched = 0;
		    if (length == patternLen) {
		      matched = (memcmp(string, pattern, (size_t)length) == 0);
		    }
		    break;
		case GLOB:
		  /*
		   * WARNING: will not work with data containing NULLs.
		   */
		    matched = Tcl_StringMatch(string, pattern);
		    break;
		case REGEXP:
		    matched = Tcl_RegExpMatchObj(interp, string, length,
						 patternObj);
		    if (matched < 0) {
			code = TCL_ERROR;
			goto done;
		    }
		    break;
	    }
	}
	if (!matched) {
	    continue;
	}

	/*
	 * We've got a match. Find a body to execute, skipping bodies
	 * that are "-".
	 */

	for (bodyIdx = i+1;  ;  bodyIdx += 2) {
	    if (bodyIdx >= switchObjc) {
		Tcl_AppendStringsToObj(Tcl_GetObjResult(interp),
			"no body specified for pattern \"", pattern,
			"\"", (char *) NULL);
		code = TCL_ERROR;
		goto done;
	    }
	    
	    if (splitObjs) {
		code = Tcl_ListObjIndex(interp, switchObjv[0], bodyIdx,
		        &bodyObj);
		if (code != TCL_OK) {
		    return code;
		}
	    } else {
		bodyObj = switchObjv[bodyIdx];
	    }
	    /*
	     * THIS FAILS IF AN OBJECT'S STRING REP HAS A NULL.
	     */
	    body = Tcl_GetStringFromObj(bodyObj, &length);
	    if ((length != 1) || (body[0] != '-')) {
		break;
	    }
	}
	code = Tcl_EvalObj(interp, bodyObj);
	if (code == TCL_ERROR) {
	    char msg[100];
	    sprintf(msg, "\n    (\"%.50s\" arm line %d)", pattern,
		    interp->errorLine);
	    Tcl_AddObjErrorInfo(interp, msg, -1);
	}
	goto done;
    }

    /*
     * Nothing matched:  return nothing.
     */

    code = TCL_OK;

    done:
    return code;
#undef EXACT
#undef GLOB
#undef REGEXP
}

#if defined(_WIN32) || defined(__WIN32__)
__declspec(dllexport)
#endif
int Nre_Init(interp)
Tcl_Interp* interp;
{
  int code;
  code = Tcl_PkgProvide(interp, (char*)"Nre", (char*)VERSION);
  if (code != TCL_OK) {
    return code;
  }

  Tcl_RegisterObjType(&tclRegExpType);
  Tcl_CreateObjCommand(interp, (char*)"nrematch",
		       (Tcl_ObjCmdProc*)Tcl_NewRegexpObjCmd,
		       NULL, NULL);
  Tcl_CreateObjCommand(interp, (char*)"nresub",
		       (Tcl_ObjCmdProc*)Tcl_NewRegsubObjCmd,
		       NULL, NULL);
  Tcl_CreateObjCommand(interp, (char*)"nredump",
		       (Tcl_ObjCmdProc*)Tcl_NewRegDumpObjCmd,
		       NULL, NULL);
  Tcl_CreateObjCommand(interp, (char*)"nrelsearch",
		       (Tcl_ObjCmdProc*)Tcl_NewLsearchObjCmd,
		       NULL, NULL);
  Tcl_CreateObjCommand(interp, (char*)"nreswitch",
		       (Tcl_ObjCmdProc*)Tcl_NewSwitchObjCmd,
		       NULL, NULL);
  return TCL_OK;
}
